/*
 * device.c: The basic device interface
 *
 * See the main source file 'vdr.c' for copyright information and
 * how to reach the author.
 *
 * $Id: device.c 1.56 2004/06/19 08:51:05 kls Exp $
 */

#include "include/vdr/device.h"
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include "include/vdr/channels.h"
#include "include/vdr/i18n.h"

// --- cDevice ---------------------------------------------------------------

// The default priority for non-primary devices:
#define DEFAULTPRIORITY  -1

int cDevice::numDevices = 0;
int cDevice::useDevice = 0;
int cDevice::nextCardIndex = 0;
int cDevice::currentChannel = 1;
cDevice *cDevice::device[MAXDEVICES] = { NULL };
cDevice *cDevice::primaryDevice = NULL;

cDevice::cDevice(void)
{
  cardIndex = nextCardIndex++;

  SetDescription("receiver on device %d", CardIndex() + 1);

  active = false;

  mute = false;

  sectionHandler = NULL;
  eitFilter = NULL;
  patFilter = NULL;
  sdtFilter = NULL;
  nitFilter = NULL;

  ciHandler = NULL;
  player = NULL;

  for (int i = 0; i < MAXRECEIVERS; i++)
      receiver[i] = NULL;

  if (numDevices < MAXDEVICES)
     device[numDevices++] = this;
  else
     esyslog("ERROR: too many devices!");
}

cDevice::~cDevice()
{
}

void cDevice::SetUseDevice(int n)
{
  if (n < MAXDEVICES)
     useDevice |= (1 << n);
}

int cDevice::NextCardIndex(int n)
{
  if (n > 0) {
     nextCardIndex += n;
     if (nextCardIndex >= MAXDEVICES)
        esyslog("ERROR: nextCardIndex too big (%d)", nextCardIndex);
     }
  else if (n < 0)
     esyslog("ERROR: illegal value in IncCardIndex(%d)", n);
  return nextCardIndex;
}

int cDevice::DeviceNumber(void) const
{
  for (int i = 0; i < numDevices; i++) {
      if (device[i] == this)
         return i;
      }
  return -1;
}

void cDevice::MakePrimaryDevice(bool On)
{
}

bool cDevice::SetPrimaryDevice(int n)
{
  n--;
  if (0 <= n && n < numDevices && device[n]) {
     isyslog("setting primary device to %d", n + 1);
     if (primaryDevice)
        primaryDevice->MakePrimaryDevice(false);
     primaryDevice = device[n];
     primaryDevice->MakePrimaryDevice(true);
     return true;
     }
  esyslog("ERROR: invalid primary device number: %d", n + 1);
  return false;
}

bool cDevice::HasDecoder(void) const
{
  return false;
}

cSpuDecoder *cDevice::GetSpuDecoder(void)
{
  return NULL;
}

cDevice *cDevice::ActualDevice(void)
{
  return NULL;
}

cDevice *cDevice::GetDevice(int Index)
{
  return NULL;
}

cDevice *cDevice::GetDevice(const cChannel *Channel, int Priority, bool *NeedsDetachReceivers)
{
  return NULL;
}

void cDevice::Shutdown(void)
{
}

bool cDevice::GrabImage(const char *FileName, bool Jpeg, int Quality, int SizeX, int SizeY)
{
  return false;
}

void cDevice::SetVideoFormat(bool VideoFormat16_9)
{
}

eVideoSystem cDevice::GetVideoSystem(void)
{
}

#define PRINTPIDS(s)

bool cDevice::HasPid(int Pid) const
{
  for (int i = 0; i < MAXPIDHANDLES; i++) {
      if (pidHandles[i].pid == Pid)
         return true;
      }
  return false;
}

bool cDevice::AddPid(int Pid, ePidType PidType)
{
  if (Pid || PidType == ptPcr) {
     int n = -1;
     int a = -1;
     if (PidType != ptPcr) { // PPID always has to be explicit
        for (int i = 0; i < MAXPIDHANDLES; i++) {
            if (i != ptPcr) {
               if (pidHandles[i].pid == Pid)
                  n = i;
               else if (a < 0 && i >= ptOther && !pidHandles[i].used)
                  a = i;
               }
            }
        }
     if (n >= 0) {
        // The Pid is already in use
        if (++pidHandles[n].used == 2 && n <= ptTeletext) {
           // It's a special PID that may have to be switched into "tap" mode
           PRINTPIDS("A");
           return SetPid(&pidHandles[n], n, true);
           }
        PRINTPIDS("a");
        return true;
        }
     else if (PidType < ptOther) {
        // The Pid is not yet in use and it is a special one
        n = PidType;
        }
     else if (a >= 0) {
        // The Pid is not yet in use and we have a free slot
        n = a;
        }
     else
        esyslog("ERROR: no free slot for PID %d", Pid);
     if (n >= 0) {
        pidHandles[n].pid = Pid;
        pidHandles[n].used = 1;
        PRINTPIDS("C");
        return SetPid(&pidHandles[n], n, true);
        }
     }
  return true;
}

void cDevice::DelPid(int Pid, ePidType PidType)
{
  if (Pid || PidType == ptPcr) {
     int n = -1;
     if (PidType == ptPcr)
        n = PidType; // PPID always has to be explicit
     else {
        for (int i = 0; i < MAXPIDHANDLES; i++) {
            if (pidHandles[i].pid == Pid) {
               n = i;
               break;
               }
            }
        }
     if (n >= 0 && pidHandles[n].used) {
        PRINTPIDS("D");
        if (--pidHandles[n].used < 2) {
           SetPid(&pidHandles[n], n, false);
           if (pidHandles[n].used == 0) {
              pidHandles[n].handle = -1;
              pidHandles[n].pid = 0;
              }
           }
        PRINTPIDS("E");
        }
     }
}

bool cDevice::SetPid(cPidHandle *Handle, int Type, bool On)
{
  return false;
}

void cDevice::StartSectionHandler(void)
{
}

int cDevice::OpenFilter(u_short Pid, u_char Tid, u_char Mask)
{
  return -1;
}

void cDevice::AttachFilter(cFilter *Filter)
{
}

void cDevice::Detach(cFilter *Filter)
{
}

bool cDevice::ProvidesSource(int Source) const
{
  return false;
}

bool cDevice::ProvidesTransponder(const cChannel *Channel) const
{
  return false;
}

bool cDevice::ProvidesChannel(const cChannel *Channel, int Priority, bool *NeedsDetachReceivers) const
{
  return false;
}

bool cDevice::SwitchChannel(const cChannel *Channel, bool LiveView)
{
  return false;
}

bool cDevice::SwitchChannel(int Direction)
{
  return false;
}

eSetChannelResult cDevice::SetChannel(const cChannel *Channel, bool LiveView)
{
}

bool cDevice::SetChannelDevice(const cChannel *Channel, bool LiveView)
{
  return false;
}

bool cDevice::HasLock(void)
{
  return true;
}

bool cDevice::HasProgramme(void)
{
  return false;
}

void cDevice::SetVolumeDevice(int Volume)
{
}

int cDevice::NumAudioTracksDevice(void) const
{
  return 0;
}

const char **cDevice::GetAudioTracksDevice(int *CurrentTrack) const
{
  return NULL;
}

void cDevice::SetAudioTrackDevice(int Index)
{
}

bool cDevice::ToggleMute(void)
{
}

void cDevice::SetVolume(int Volume, bool Absolute)
{
}

int cDevice::NumAudioTracks(void) const
{
  return 0;
}

const char **cDevice::GetAudioTracks(int *CurrentTrack) const
{
  return NULL;
}

void cDevice::SetAudioTrack(int Index)
{
}

bool cDevice::CanReplay(void) const
{
  return HasDecoder();
}

bool cDevice::SetPlayMode(ePlayMode PlayMode)
{
  return false;
}

int64_t cDevice::GetSTC(void)
{
  return -1;
}

void cDevice::TrickSpeed(int Speed)
{
}

void cDevice::Clear(void)
{
}

void cDevice::Play(void)
{
}

void cDevice::Freeze(void)
{
}

void cDevice::Mute(void)
{
}

void cDevice::StillPicture(const uchar *Data, int Length)
{
}

bool cDevice::Replaying(void) const
{
  return player != NULL;
}

bool cDevice::AttachPlayer(cPlayer *Player)
{
  return false;
}

void cDevice::Detach(cPlayer *Player)
{
}

void cDevice::StopReplay(void)
{
}

bool cDevice::Poll(cPoller &Poller, int TimeoutMs)
{
  return false;
}

bool cDevice::Flush(int TimeoutMs)
{
  return true;
}

int cDevice::PlayVideo(const uchar *Data, int Length)
{
  return -1;
}

void cDevice::PlayAudio(const uchar *Data, int Length)
{
}

int cDevice::Ca(void) const
{
  return 0;
}

int cDevice::Priority(void) const
{
  return 0;
}

int cDevice::CanShift(int Ca, int Priority, int UsedCards) const
{
  return -1;//XXX+ too complex with multiple recordings per device
}

int cDevice::ProvidesCa(const cChannel *Channel) const
{
  int Ca = Channel->Ca();
  if (Ca == CardIndex() + 1)
     return 1; // exactly _this_ card was requested
  if (Ca && Ca <= MAXDEVICES)
     return 0; // a specific card was requested, but not _this_ one
  return !Ca; // by default every card can provide FTA
}

bool cDevice::Receiving(bool CheckAny) const
{
  return false;
}

void cDevice::Action(void)
{
}

bool cDevice::OpenDvr(void)
{
  return false;
}

void cDevice::CloseDvr(void)
{
}

bool cDevice::GetTSPacket(uchar *&Data)
{
  return false;
}

bool cDevice::AttachReceiver(cReceiver *Receiver)
{
  return false;
}

void cDevice::Detach(cReceiver *Receiver)
{
}

// --- cTSBuffer -------------------------------------------------------------

cTSBuffer::cTSBuffer(int File, int Size, int CardIndex)
{
  f = File;
  size = Size / TS_SIZE * TS_SIZE;
  cardIndex = CardIndex;
  tsRead = tsWrite = 0;
  buf = (f >= 0 && size >= TS_SIZE) ? MALLOC(uchar, size + TS_SIZE) : NULL;
  // the '+ TS_SIZE' allocates some extra space for handling packets that got split by a buffer roll-over
  firstRead = true;
}

cTSBuffer::~cTSBuffer()
{
  free(buf);
}

int cTSBuffer::Read(void)
{
  if (buf) {
     cPoller Poller(f, false);
     bool repeat;
     int total = 0;
     do {
        repeat = false;
        if (firstRead || Used() > TS_SIZE || Poller.Poll(100)) { // only wait if there's not enough data in the buffer
           firstRead = false;
           if (tsRead == tsWrite)
              tsRead = tsWrite = 0; // keep the maximum buffer space available
           if (tsWrite >= size && tsRead > 0)
              tsWrite = 0;
           int free = tsRead <= tsWrite ? size - tsWrite : tsRead - tsWrite - 1;
           if (free > 0) {
              int r = read(f, buf + tsWrite, free);
              if (r > 0) {
                 total += r;
                 tsWrite += r;
                 if (tsWrite >= size && tsRead > 0) {
                    tsWrite = 0;
                    repeat = true; // read again after a boundary roll-over
                    }
                 }
              }
           }
        } while (repeat);
     return total;
     }
  return -1;
}

uchar *cTSBuffer::Get(void)
{
  if (Used() >= TS_SIZE) {
     uchar *p = buf + tsRead;
     if (*p != TS_SYNC_BYTE) {
        esyslog("ERROR: not sync'ed to TS packet on device %d", cardIndex);
        int tsMax = tsRead < tsWrite ? tsWrite : size;
        for (int i = tsRead; i < tsMax; i++) {
            if (buf[i] == TS_SYNC_BYTE) {
               esyslog("ERROR: skipped %d bytes to sync on TS packet on device %d", i - tsRead, cardIndex);
               tsRead = i;
               return NULL;
               }
            }
        if ((tsRead = tsMax) >= size)
           tsRead = 0;
        return NULL;
        }
     if (tsRead + TS_SIZE > size) {
        // the packet rolled over the buffer boundary, so let's fetch the rest from the beginning (which MUST be there, since Used() >= TS_SIZE)
        int rest = TS_SIZE - (size - tsRead);
        memcpy(buf + size, buf, rest);
        tsRead = rest;
        }
     else if ((tsRead += TS_SIZE) >= size)
        tsRead = 0;
     return p;
     }
  return NULL;
}
