#include "include/vdr/channels.h"

int cChannel::Transponder() const {return sid;}
int cChannel::Transponder(int Frequency, char Polarization)
{
  // some satellites have transponders at the same frequency, just with different polarization:
  switch (tolower(Polarization)) {
    case 'h': Frequency += 100000; break;
    case 'v': Frequency += 200000; break;
    case 'l': Frequency += 300000; break;
    case 'r': Frequency += 400000; break;
    }
  return Frequency;
}
cChannel* cChannels::GetByNumber(int a, int b) {
  cChannel *ch = new cChannel;
  ch->SetId(0,0,a);
  return ch;
}
void cChannel::SetId(int Nid, int Tid, int Sid, int Rid) {
  sid = Sid;
  tid = Tid;
}
void cChannel::SetPids(int Vpid, int Ppid, int *Apids, char ALangs[][4], int *Dpids, char DLangs[][4], int Tpid)
{
  vpid = Vpid;
  ppid = Ppid;
  tpid = Tpid;
  for (int i = 0; i < MAXAPIDS; i++)
      apids[i] = Apids[i];
  apids[MAXAPIDS] = 0;
  for (int i = 0; i < MAXDPIDS; i++) {
     dpids[i] = Dpids[i];
  }
  dpids[MAXDPIDS] = 0;
}

cChannel::cChannel() {
  caids[0]=101;
  caids[1]=0;
  source=1;
  sid=1;
}
cChannel::~cChannel() {
}

void parseHdr(cChannel *ch, int *hdr, int hdrlen)
{
    int vpid = 0, ppid = 0, tpid = 0, acnt = 0, dcnt = 0;
    int apid[MAXAPIDS], dpid[MAXDPIDS];
    memset(apid, 0, sizeof(int)*MAXAPIDS);
    memset(dpid, 0, sizeof(int)*MAXDPIDS);
    ch->SetId(0, hdr[1], hdr[0], 0);
    for(int i = 2; i < hdrlen; i+=2) {
        switch(hdr[i+1]) {
            case 0: //Audio
              apid[acnt++] = hdr[i];
              break;
            case 1: //Video
              vpid = hdr[i];
              break;
            case 2: //PCR
              ppid = hdr[i];
              break;
            case 3: //Teletext
              tpid = hdr[i];
              break;
            case 5: //Other
              dpid[dcnt++] = hdr[i];
              break;
        }
    }
    ch->SetPids(vpid, ppid, apid, NULL, dpid, NULL, tpid);
}

extern uchar pmtbuf[4096];
extern int pmtlen;
int GetCaDescriptors(int Source, int Transponder, int ServiceId, const unsigned short *CaSystemIds, int BufSize, uchar *Data, bool &StreamFlag) {
  memcpy(Data, pmtbuf,pmtlen);
//  printf("Transfering %d bytes\n", pmtlen);
  return pmtlen;
}

extern int csafd;
void _SetCaDescr(ca_descr_t *ca_descr) {
  //fprintf(stderr, "Sending key(%d) %c %d %x%x%x%x%x%x%x%x\n", csafd,
  //                (ca_descr->parity ? 'O' : 'E'), ca_descr->index,
  //                ca_descr->cw[0], ca_descr->cw[1], ca_descr->cw[2],
  //                ca_descr->cw[3], ca_descr->cw[4], ca_descr->cw[5],
  //                ca_descr->cw[6], ca_descr->cw[7]);
  if (csafd<0)
    return;
  if  (ca_descr->parity==0) // even key
    write(csafd,"E",1);
  else // odd key
    write(csafd,"O",1);
  write(csafd,&ca_descr->index,sizeof(ca_descr->index));
  write(csafd,ca_descr->cw,8);
}

void cDevice::SetCaDescr(ca_descr_t *ca_descr) {
  _SetCaDescr(ca_descr);
}

void _SetCaPid(ca_pid_t *ca_pid) {
  //fprintf(stderr, "Sending key P %d PID: %d\n", ca_pid->index, ca_pid->pid);
  if (csafd < 0)
    return;
  write(csafd,"P",1);
  write(csafd,&ca_pid->index,sizeof(ca_pid->index));
  write(csafd,&ca_pid->pid,sizeof(ca_pid->pid));
}

void cDevice::SetCaPid(ca_pid_t *ca_pid) {
  _SetCaPid(ca_pid);
}

void _ForceIndex(int index) {
  //fprintf(stderr, "Sending key F %d\n", index);
  if (csafd < 0)
    return;
  write(csafd,"F",1);
  write(csafd,&index,sizeof(index));
}

#define DEV_DVB_ADAPTER "/dev/dvb/adapter"
#ifndef PATH_MAX
#define PATH_MAX        512
#endif
static const char *DvbName(const char *Name, int n)
{
  static char buffer[PATH_MAX];
  snprintf(buffer, sizeof(buffer), "%s%d/%s%d", DEV_DVB_ADAPTER, n, Name, 0);
  return buffer;
}

int DvbOpen(const char *Name, int n, int Mode, bool ReportError)
{
  const char *FileName=DvbName(Name,n);
  int fd=open(FileName,Mode);
  if(fd<0 && ReportError) LOG_ERROR_STR(FileName);
  return fd;
}
