#include "othercamd.h"

#include "include/vdr/plugin.h"
#include "include/vdr/menuitems.h"
#include "include/vdr/dvbdevice.h"
#include "include/vdr/status.h"
#include "include/vdr/sclink.h"
#include "include/vdr/channels.h"
#include "include/vdr/config.h"
#include <dlfcn.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

static char scDir[256], libDir[256];

const char *cPlugin::ConfigDirectory(const char *PluginName) {return scDir;}
extern "C" void *VDRPluginCreator(void);
int DvbOpen(const char *Name, int n, int Mode, bool ReportError=false);
void parseHdr(cChannel *ch, int *hdr, int hdrlen);

void help (char *prg) {
  printf("%s --lib <directory> --dir <directory> --cardnum <card number>\n", prg);
  exit(-1);
}

uchar pmtbuf[4096];
int  pmtlen;
int csafd = -1, camfd=-1;

bool init_sc(int cardnum, cPlugin **plugin, cDvbDevice **dev) {
  if(csafd > 0)
    camfd=DvbOpen("ca",cardnum,O_RDWR|O_NONBLOCK);

  *dev = new cDvbDevice(cardnum);

  char *fileName;
  asprintf(&fileName,"%s/libvdr-sc.so.%s",libDir,APIVERSION);
  void *handle;
  handle = dlopen(fileName, RTLD_NOW);
  free(fileName);
  if(!handle) {
    printf("dload failed: %s\n",dlerror());
    return false;
    }
  void *(*creator)(void);
  creator = (void *(*)(void))dlsym(handle, "VDRPluginCreator");
  if(!creator) {
    printf("dlsym failed: %s\n",dlerror());
    exit(1);
    }
  *plugin=(cPlugin *)creator();
  cPlugin *sc = *plugin;

  printf("sasc (%d):  initializing plugin: SoftCam (%s): %s\n", cardnum, sc->Version(), sc->Description());
  if (!sc->Initialize()) {
    printf("sasc (%d): Failed\n", cardnum);
    return false;
  }
  printf("sasc (%d): starting plugin:\n",cardnum);
  if (!sc->Start()) {
    printf("sasc (%d): Failed\n", cardnum);
    return false;
  }

  char scCap[2];
  snprintf(scCap,2,"%d",cardnum+1);
  sc->SetupParse("ScCaps",scCap);
  sc->SetupParse("LoggerActive","2");
  return true;
}


int main(int argc, char *argv[]) {
  cPlugin *sc;
  cDvbDevice *dev;

  int ret = 0;
  int cardnum = -1;
  int use_evo = 0;
  int evopid = 0;
  int infd=fileno(stdin);
  {
    scDir[0]=0; libDir[0]=0;
    static struct option Long_Options[] = {
      {"help", 0, 0, 'h'},
      {"dir", 1, 0, 'd'},
      {"cardnum", 1, 0, 'c'},
      {"fifo", 1, 0, 'f'},
      {"writekey", 1, 0, 'w'},
      {"lib", 1, 0, 'l'},
      {"useevo", 0, 0, 'e'},
      {0, 0, 0, 0}
    };
    int c;
    int Option_Index = 0;

    while (1) {
      c = getopt_long(argc, argv, "c:d:ef:hl:w:",
                      Long_Options, &Option_Index);
      if (c == EOF)
        break;
      switch (c) {
        case 'c':
          cardnum=atoi(optarg);
          break;
        case 'd':
          strncpy(scDir,optarg,255);
          break;
        case 'e':
          use_evo = 1;
          break;
        case 'f':
          infd=open(optarg, O_RDONLY);
          if(infd <0) {
            fprintf(stderr, "sasc failed to open input file '%s'\n", optarg);
            return(-1);
          }
          break;
        case 'h':
          help(argv[0]);
          break;
        case 'l':
          strncpy(libDir,optarg,255);
          break;
        case 'w':
          csafd=open(optarg, O_WRONLY);
          printf("opening keyfile: %s -> %d\n",optarg, csafd);
          break;
      }
    }
    if (cardnum == -1 || scDir[0] == 0) {
      fprintf(stderr, "Incorrect options specified\n");
      help(argv[0]);
    }
    if (libDir[0] == 0) {
      snprintf(libDir, 256, "%s/lib", scDir);
    }
    fprintf(stderr,"Set dir='%s',cardnum=%d\n",scDir,cardnum);
    fprintf(stderr,"Dynamic libraries in '%s'\n", libDir);
  }
  pmtlen = 0;

  //If 'write' was specified, then we will grab the CA (if available)
  //to fool SC into thinking this is a budget card

  if(! use_evo) {
    init_sc(cardnum,&sc,&dev);
  } else {
    if(! init_evocamd(cardnum, scDir)) {
      shutdown_evocamd(cardnum);
      return(-1);
    }
    evopid = start_loopbackca(cardnum);
    printf("sasc(%d) - EvoCamdPID = %d\n", cardnum, evopid);
  }
  int chan_count=0;
  struct pollfd ufs;
  memset(&ufs,0, sizeof(ufs));
  ufs.fd=infd;
  ufs.events = POLLIN;
  int lastchan = -1;
  int oldhdr[34], oldhdrlen = 0;
  while(1) {
    uchar buf[4096], *ptr;
    int hdr[34];
    int len,size,hdrlen, i;
    if(1 || ufs.revents==POLLIN) {
      size=read(ufs.fd, &hdrlen, sizeof(int));
      if(0 == size) {
        ret = 0;
        break;
      }
      if(size != sizeof(int)) {
        fprintf(stderr, "sasc (%d): Fatal error:  didn't read header length.\n", cardnum);
        perror("");
        ret = -1;
        break;
      }
      size=read(ufs.fd, &hdr, sizeof(int) * hdrlen);
      if(size != sizeof(int)*hdrlen) {
        printf("sasc (%d): Fatal error:  didn't read header\n", cardnum);
        perror("");
        ret = -1;
        break;
      }
      size=read(ufs.fd, &len, sizeof(int));
      if(size != sizeof(int)) {
        printf("sasc (%d): Fatal error:  didn't read length\n", cardnum);
        perror("");
        ret = -1;
        break;
      }
      size=read(ufs.fd, buf, len);
      if (size != len) {
        printf("sasc (%d): Fatal error: couldnt read data\n", cardnum);
        perror("");
        ret = -1;
        break;
      }
      printf("sasc (%d): got PMT (len = %d)",cardnum, len);
      for(i=0;i<len;i++) {
        printf("%02x ",buf[i]);
      }
      printf("\n");
      ptr=buf;

      while( (len > 1) && !( (*ptr == 0x09)&&(*(ptr+1) == 0x0f || *(ptr+1) == 0x04))) {
        ptr++;
        len--;
      }

      printf("sasc (%d): corrected to (len = %d)\n",cardnum, len);
      bool skip = false;
      if (len == 0)
        skip = true;
      else if(!use_evo && len == pmtlen && memcmp(pmtbuf, ptr, len) == 0 &&
              hdrlen == oldhdrlen &&  memcmp(hdr, oldhdr, hdrlen) == 0) {
        skip=true;
      }
      if(!skip) {
        struct ScLink link;
        cChannel *ch;

        printf("sasc (%d): Changing channels: %d %d %d\n",cardnum, len, pmtlen, i);
        oldhdrlen = hdrlen;
        memcpy(oldhdr, hdr, hdrlen);
        memcpy(pmtbuf,ptr,len);
        pmtlen=len;
        if (use_evo) {
          send_evocamd(cardnum, hdr[0], 1);
        } else {
          if(Channels.First())
            Channels.Del(Channels.First());
  
          ch = new cChannel();
          parseHdr(ch, hdr, hdrlen);
          Channels.Add(ch);
          PrepareScLink(&link,dev,OP_TUNE);
          link.data.tune.source=ch->Source();
          link.data.tune.transponder=ch->Transponder();
          DoScLinkOp(sc, &link);
  
          for (i=2; i < hdrlen; i+=2) {
            PrepareScLink(&link,dev,OP_ADDPID);
            link.data.pids.pid=hdr[i];
            link.data.pids.type=hdr[i+1]; //type == "OTHER"
            DoScLinkOp(sc, &link);
          }
          lastchan=hdr[0];
       }
      } else if(len != 0 && csafd != -1) {
        // ths shoud have a mutex to be safe!
        //printf("sasc (%d): Resending  channel %d\n", hdr[0]);
        //write(csafd, "B", 1);
      }
    }
  }
  if(evopid > 0) {
    kill(evopid, SIGTERM);
    waitpid(evopid, NULL, 0);
  }
  shutdown_evocamd(cardnum);
  return(ret);
}

